/* 
 *    fame.c
 *
 *	Copyright (C) Vivien Chappelier - 2000-2001
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#ifdef HAS_LAME
#include <lame/lame.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<sys/ioctl.h>
#include	<termio.h>	/* for ioctl functions */

struct termio	ostate;		/* saved tty state */
struct termio	nstate;		/* values for editor mode */

#include <fame.h>

#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <arpa/inet.h>

#include "types.h"
#include "dsp.h"
#include "v4l.h"

#define DEFAULT_BUFFER_SIZE 1024*1024
#define CHUNK_SIZE 1024

fuint32_t video_frame;
fuint32_t audio_frame;
fuint32_t sync_frame, start_frame, ev_frame, ea_frame;

volatile fbool_t quit, pause_record;
char *profile = "mpeg1";
char *motion = "pmvfast";
char video_device[MAXPATHLEN] = "/dev/video\0";
pthread_t video_thread, audio_thread;
pthread_mutex_t buffer_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t buffer_cond = PTHREAD_COND_INITIALIZER;

fame_parameters_t fp = FAME_PARAMETERS_INITIALIZER;
int buffer_size = DEFAULT_BUFFER_SIZE;
int video_fd = 1;
int audio_fd = 0;
int audiobitrate = 64;

static fbool_t init();
static fbool_t cleanup();
static void *audio_callback(void *args);
static void *video_callback(void *args);
static void break_handler(int signal);

/*
 * parse command line
 */
int parse(int argc, char *argv[])
{
  int index = -1;
  int c;
  /* options */
#ifdef _GNU_SOURCE
  const struct option longopts[]={
    {"picture",1,0,'p'}, /* picture size: widthxheight or CIF,SIF,QCIF */
    {"coding",1,0,'c'},  /* coding sequence: for example "IBBPBBPBB"   */
    {"quality",1,0,'q'}, /* quality: percentage of video quality */
    {"bitrate",1,0,'b'}, /* bitrate: in bits/s, K for kilo, M for mega */
    {"audiorate",1,0,'u'},/* audiobitrate: in bits/s, K for kile, M for mega */
    {"buffer",1,0,'B'},  /* buffer size: in bytes, K for kilo, M for mega */
    {"alpha",1,0,'a'},   /* shape quality: percentage of shape accuracy */
    {"search",1,0,'s'},  /* search range: motion estimation range */
    {"fps",1,0,'f'},     /* framerate: fractional number of frames/sec */
    {"refresh",1,0,'r'}, /* refresh: number of frames before a new sequence */
    {"options",1,0,'o'}, /* options: intra_only, lossless_shape */
    {"profile",1,0,'P'}, /* profile name: name of the profile to use */
    {"motion",1,0,'M'},  /* motion: motion estimation algorithm */
    {"device",1,0,'d'},  /* video device */
    {"verbose",0,0,'v'}, /* verbose: when set, print messages */
    {"help",0,0,'h'},    /* help: display help message */
    {0,0,0,0}
  };
#endif
  const char * shortopts = "p:c:d:q:b:B:a:s:f:r:o:P:M:u:vh";

#ifdef _GNU_SOURCE
  while((c = getopt_long(argc,argv,shortopts,longopts,&index)) != -1)
#else
  while((c = getopt(argc,argv,shortopts)) != -1)
#endif
  {
    switch(c) 
    {
      case 'p':
	/* picture size: widthxheight or CIF,SIF,QCIF,QSIF */
	if(!strcmp("CIF", optarg)) {
	  fp.width = 352;
	  fp.height = 288;
	} else 
	if(!strcmp("SIF", optarg)) {
	  fp.width = 352;
	  fp.height = 240;
	} else
	if(!strcmp("QCIF", optarg)) {
	  fp.width = 176;
	  fp.height = 144;
	} else 
	if(strchr(optarg, 'x')) {
	  fp.width = atoi(optarg);
	  fp.height = atoi(strchr(optarg, 'x')+1);
	} else {
	  return(index);
	}
      break;

      /* coding sequence: for example "IBBPBBPBB"   */
      case 'c':
	fp.coding = optarg;
      break;

      /* coding sequence: for example "IBBPBBPBB"   */
      case 'd':
	strcpy(video_device, optarg);
      break;


      /* quality percentage */
      case 'q':
	fp.quality = atoi(optarg);
      break;

      /* bitrate: in bits/s, K for kilo, M for mega */
      case 'b':
	fp.bitrate = atoi(optarg);
	if(optarg[strlen(optarg)-1] == 'K' ||
	   optarg[strlen(optarg)-1] == 'k')
	  fp.bitrate *= 1024;
	if(optarg[strlen(optarg)-1] == 'M' ||
	   optarg[strlen(optarg)-1] == 'm')
	  fp.bitrate *= 1024*1024;
      break;

      /* bitrate of audio stream */
      case 'u':
        audiobitrate = atoi(optarg);
	if(optarg[strlen(optarg)-1] == 'K' ||
	   optarg[strlen(optarg)-1] == 'k')
	  audiobitrate *= 1024;
	if(optarg[strlen(optarg)-1] == 'M' ||
	   optarg[strlen(optarg)-1] == 'm')
	  audiobitrate *= 1024*1024;
      break;

      /* buffer size: in bytes, K for kilo, M for mega */
      case 'B':
	buffer_size = atoi(optarg);
	if(optarg[strlen(optarg)-1] == 'K' ||
	   optarg[strlen(optarg)-1] == 'k')
	  buffer_size *= 1024;
	if(optarg[strlen(optarg)-1] == 'M' ||
	   optarg[strlen(optarg)-1] == 'm')
	  buffer_size *= 1024*1024;
      break;

      /* shape quality percentage */
      case 'a':
	fp.shape_quality = atoi(optarg);
      break;

      /* search range: motion estimation range */
      case 's':
	fp.search_range = atoi(optarg);
      break;

      /* framerate: fractional number of frames/sec */
      case 'f':
	fp.frame_rate_num = atoi(optarg);
	if(strchr(optarg, '/') != NULL)
	  fp.frame_rate_den = atoi(strchr(optarg, '/') + 1);
	else
	  fp.frame_rate_den = 1;
      break;

      /* refresh: number of frames before a new sequence */
      case 'r':
	fp.frames_per_sequence = atoi(optarg);
      break;

      /* profile name: name of the profile to use */
      case 'P':
	profile = optarg;
      break;

      /* motion: motion estimation algorithm */
      case 'M':
	motion = optarg;
      break;

      /* verbose: when set, print messages */
      case 'v':
	fp.verbose = 1;
      break;

      /* help */
      case 'h':
	return(-1);
      break;

      default: 
	return(index);
    }
  }

  /* force at least one argument */
  if(argc < optind+1)
    return(-1);

  /* output file */
  if(argc >= optind+1) {
    /* use stdout */
    if(!strncmp(argv[optind+0], "-", strlen("-")))
    {
      video_fd = 1;
    } else
    /* use udp */
    if(!strncmp(argv[optind+0], "udp://", strlen("udp://")))
    {
      char * host;
      int port;
      int sock;
      
      port = 10000;
      host = argv[optind+0] + strlen("udp://");
      if(strchr(host, ':') != NULL) {
	port = atoi(strchr(host, ':') + 1);
	*strchr(host, ':') = 0;
      }
      video_fd = udp_open(host, port);
    } else
      video_fd = open(argv[optind+0], O_WRONLY | O_CREAT | O_TRUNC, 0666);

    if(video_fd < 0) {
      char error[256];
      sprintf(error, "Error opening %s for writing", argv[optind+0]);
      perror(error);
      exit(1);
    }
  }

  /* audio file */
  if(argc >= optind+2) {
#ifndef HAS_LAME
    fprintf(stderr, "Error, audio sound support not compiled\n");
    exit(1);
#endif
    /* use stdout */
    if(!strncmp(argv[optind+0], "-", strlen("-")))
    {
      video_fd = 1;
    } else
    /* use udp */
    if(!strncmp(argv[optind+1], "udp://", strlen("udp://")))
    {
      char * host;
      int port;
      int sock;
      
      port = 10001;
      host = argv[optind+1] + strlen("udp://");
      if(strchr(host, ':') != NULL) {
	port = atoi(strchr(host, ':') + 1);
	*strchr(host, ':') = 0;
      }
      audio_fd = udp_open(host, port);
    } else
      audio_fd = open(argv[optind+1], O_WRONLY | O_CREAT | O_TRUNC, 0666);

    if(audio_fd < 0) {
      char error[256];
      sprintf(error, "Error opening %s for writing", argv[optind+1]);
      perror(error);
      exit(1);
    }
  }

  return(0);
}

void usage(char const *name)
{
  fprintf(stderr, "usage: %s options video [audio]\n", name);
  fprintf(stderr, "options are\n");
  fprintf(stderr, "\t%s\t%s\n",
	  "-p --picture",
	  "picture size: widthxheight or CIF,SIF,QCIF");
  fprintf(stderr, "\t%s\t%s\n",
	  "-c --coding",
	  "coding sequence: for example \"IPPPP\"");
  fprintf(stderr, "\t%s\t%s\n",
	  "-d --device",
	  "video device: for example \"/dev/video0\"");
  fprintf(stderr, "\t%s\t%s\n",
	  "-q --quality",
	  "quality: percentage of video quality");
  fprintf(stderr, "\t%s\t%s\n",
          "-b --bitrate",
	  "bitrate: video bitrate in bits/s [K kilo, M mega]");
  fprintf(stderr, "\t%s\t%s\n",
          "-u --audiorate",
 	  "audiobitrate: mp3 bitrate in bits/s [K kilo, M mega]");
  fprintf(stderr, "\t%s\t%s\n",
          "-B --buffer",
	  "buffer size: in bytes [K kilo, M mega]");
  fprintf(stderr, "\t%s\t%s\n",
          "-a --alpha",
	  "shape quality: percentage of shape accuracy");
  fprintf(stderr, "\t%s\t%s\n",
	  "-s --search",
	  "search range: motion estimation range");
  fprintf(stderr, "\t%s\t%s\n",
	  "-f --fps",
	  "framerate: fractional number of frames/sec");
  fprintf(stderr, "\t%s\t%s\n",
	  "-r --refresh",
	  "number of frames before a new sequence");
  fprintf(stderr, "\t%s\t%s\n",
	  "-P --profile",
	  "profile: name of the profile to use");
  fprintf(stderr, "\t%s\t%s\n",
	  "-M --motion",
	  "motion: motion estimation algorithm");
  fprintf(stderr, "\t%s\t%s\n",
	  "-v --verbose",
	  "verbose: when set, print messages");
  fprintf(stderr, "\t%s\t%s\n",
	  "-h --help",
	  "display this message");

  exit(1);
}

static void *audio_callback(void *args)
{
#ifdef HAS_LAME
  lame_global_flags * lc;
  input_dsp_context_t * dsp;
  short int     *pcm;
  unsigned char *buffer;
  int size;

  fd_set  readfds;
  struct  timeval timeout;

  pcm = (short int *) malloc(3528*sizeof(short int));
  buffer = (unsigned char *) malloc(DEFAULT_BUFFER_SIZE*4);
  lc = lame_init();
  lame_set_out_samplerate(lc, 44100);
  lame_set_brate(lc, audiobitrate/1000); /* bitrate is in Kbits/s in lame */
  lame_init_params(lc);
  if(!(dsp = dsp_init())) {
    quit = true;
    fprintf(stderr, "audio: could not open dsp\n");
    pthread_cond_signal(&buffer_cond);
    return(NULL);
  }

  while(!quit) {
    dsp_read(dsp, pcm, 1764*sizeof(short int));

    if (start_frame && (ea_frame < start_frame))
    {
    	ea_frame ++;
    	continue;
    }
    if (!pause_record || (audio_frame < sync_frame))
    {
    	if (pause_record && (audio_frame < sync_frame))
    	    fprintf (stderr, "\rAudio sync\n");
        size = lame_encode_buffer(lc, pcm, pcm, 1764, buffer, 0);
        split_write (audio_fd, buffer, size);
        audio_frame++;
	/*fprintf (stderr, "\r                                               frame = %d / %d \r", audio_frame, video_frame);*/
    } else
    	ea_frame++;

    memset (&readfds, 0, sizeof(fd_set));
    FD_SET (0, &readfds);
    timeout.tv_sec = 0L;
    timeout.tv_usec = 0L;
    if (select (1, &readfds, (fd_set *) NULL, (fd_set *) NULL, &timeout))
    {
    	char qq = fgetc (stdin);
    	if ((qq == 'q') || (qq == 'Q'))
    	    quit = true;
    	if (pause_record == true)
    	{
    	    pause_record = false;
        } else 
            pause_record = true;
        sync_frame = fame_max (audio_frame, video_frame);
        start_frame = fame_max (ea_frame, ev_frame);
    }
  }
  size = lame_encode_flush(lc, buffer, 0);
  split_write(audio_fd, buffer, size);
  lame_close(lc);
  free(buffer);
  free(pcm);
  dsp_close(dsp);
#endif
  return(NULL);
}

static void *video_callback(void *args)
{
  fame_context_t *fc;
  fame_object_t *object;
  fame_yuv_t yuv;
  input_v4l_context_t * v4l;
  unsigned char *buffer;
  int size;
  fd_set  readfds;
  struct  timeval timeout;
                

  buffer = (unsigned char *) malloc(DEFAULT_BUFFER_SIZE);
  yuv.w = fp.width;
  yuv.h = fp.height;
  yuv.p = fp.width;
  yuv.y = (unsigned char *) malloc(fp.width*fp.height*12/8);
  yuv.u = yuv.y + fp.width*fp.height;
  yuv.v = yuv.u + fp.width*fp.height/4;
  if(!(v4l = v4l_init(fp.width, fp.height))) {
    quit = true;
    fprintf(stderr, "video: could not open video4linux device\n");
    pthread_cond_signal(&buffer_cond);
    return(NULL);
  }
  if(!(fc = fame_open())) {
    quit = true;
    fprintf(stderr, "video: error initializing libfame\n");
    pthread_cond_signal(&buffer_cond);
    return(NULL);
  }

  object = NULL;
  if(!strcmp(profile, "mpeg1") || !strcmp(profile, "MPEG1"))
    object = fame_get_object(fc, "profile/mpeg1");
  if(!strcmp(profile, "mpeg4") || !strcmp(profile, "MPEG4"))
    object = fame_get_object(fc, "profile/mpeg4");
  if(!strcmp(profile, "mpeg4_simple") || !strcmp(profile, "MPEG4_SIMPLE")) 
    object = fame_get_object(fc, "profile/mpeg4/simple");
  if(!strcmp(profile, "mpeg4_shape") || !strcmp(profile, "MPEG4_SHAPE")) 
    object = fame_get_object(fc, "profile/mpeg4/shape");
  if(object)
    fame_register(fc, "profile", object);
  else
    fprintf(stderr, "unknown profile %s\n", profile);

  object = NULL;
  if(!strcmp(motion, "none") || !strcmp(motion, "NONE"))
    object = fame_get_object(fc, "motion/none");
  if(!strcmp(motion, "pmvfast") || !strcmp(motion, "PMVFAST"))
    object = fame_get_object(fc, "motion/pmvfast");
  if(!strcmp(motion, "fourstep") || !strcmp(motion, "FOURSTEP")) 
    object = fame_get_object(fc, "motion/fourstep");
  if(object)
    fame_register(fc, "motion", object);
  else
    fprintf(stderr, "unknown motion estimation algorithm %s\n", motion);

  fame_init(fc, &fp, buffer, DEFAULT_BUFFER_SIZE);

  while(!quit) {
    v4l_read(v4l, yuv.y, fp.width*fp.height*12/8);

    if (start_frame && (ev_frame < start_frame))
    {
    	ev_frame ++;
    	continue;
    }
    if (!pause_record || (video_frame < sync_frame))
    {
        if (pause_record && (video_frame < sync_frame))
            fprintf (stderr, "\rVideo sync\n");
    	if (video_frame % fp.frame_rate_den == 0) {
	  fame_start_frame(fc, &yuv, NULL);
	  while(size = fame_encode_slice(fc))
	    split_write(video_fd, buffer, size);
	  fame_end_frame(fc, NULL);
        }
        video_frame++;
    } else
    	ev_frame++;

    memset (&readfds, 0, sizeof(fd_set));
    FD_SET (0, &readfds);
    timeout.tv_sec = 0L;
    timeout.tv_usec = 0L;
    if (select (1, &readfds, (fd_set *) NULL, (fd_set *) NULL, &timeout))
    {
    	char qq = fgetc (stdin);
    	if ((qq == 'q') || (qq == 'Q'))
    	    quit = true;
    	if (pause_record == true)
    	{
    	    pause_record = false;
        } else 
            pause_record = true;
        sync_frame = fame_max (audio_frame, video_frame);
        start_frame = fame_max (ea_frame, ev_frame);
    }
  }

  size = fame_close(fc);
  split_write(video_fd, buffer, size);
  free(buffer);
  free(yuv.y);
  v4l_close(v4l);
  return(NULL);
}

/*
 * write to output by packets of CHUNK_SIZE length
 */
int split_write(int fd, unsigned char *data, unsigned int size)
{
    fd_set set;
    int r, w;

    pthread_mutex_lock(&buffer_mutex);
    w = 0;
    while(size > CHUNK_SIZE) {
        r = write(fd, data, CHUNK_SIZE);
	if(r < 0) {
	  pthread_mutex_unlock(&buffer_mutex);
	  pthread_cond_signal(&buffer_cond);
	  return(r);
	}
	w += r;
        size -= CHUNK_SIZE;
        data += CHUNK_SIZE;
      
        FD_ZERO(&set);
        FD_SET(fd, &set);
        if(select(fd+1, NULL, &set, NULL, NULL) <= 0) break;
    }
    r = write(fd, data, size);
    if(r < 0) {
      pthread_mutex_unlock(&buffer_mutex);
      pthread_cond_signal(&buffer_cond);
      return(r);
    }
    w += r;
    pthread_mutex_unlock(&buffer_mutex);
    pthread_cond_signal(&buffer_cond);
    return(w);
}

/*
 * open UDP socket
 */

int udp_open(char * address, int port)
{
  int enable = 1L;
  struct sockaddr_in stAddr;
  struct sockaddr_in stLclAddr;
  struct hostent * host;
  int sock;
  
  stAddr.sin_family = AF_INET;
  stAddr.sin_port = htons(port);
  if((host = gethostbyname(address)) == NULL) return(0);
  stAddr.sin_addr = *((struct in_addr *) host->h_addr_list[0]);

  /* Create a UDP socket */
  if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
        return(0);

  /* Allow multiple instance of the client to share the same address and port */  if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &enable, sizeof(unsigned long int)) < 0) return(0);

#ifdef USE_MULTICAST
  /* If the address is multicast, register to the multicast group */
  if(is_address_multicast(stAddr.sin_addr.s_addr))
  {
    struct ip_mreq stMreq;
  
    /* Bind the socket to port */
    stLclAddr.sin_family      = AF_INET;
    stLclAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    stLclAddr.sin_port        = stAddr.sin_port;
    if(bind(sock, (struct sockaddr*) & stLclAddr, sizeof(stLclAddr)) < 0) return(0);

    /* Register to a multicast address */
    stMreq.imr_multiaddr.s_addr = stAddr.sin_addr.s_addr;
    stMreq.imr_interface.s_addr = INADDR_ANY;
    if(setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) & stMreq, sizeof(stMreq)) < 0) return(0);
  }
  else
#endif
  {
    /* Bind the socket to port */
    stLclAddr.sin_family      = AF_INET;
    stLclAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    stLclAddr.sin_port        = htons(0);
    if(bind(sock, (struct sockaddr*) & stLclAddr, sizeof(stLclAddr)) < 0)
        return(0);
  }

  connect(sock, (struct sockaddr*) & stAddr, sizeof(stAddr));

  return(sock);
}

fbool_t init()
{
  quit = false;
  pause_record = false;

  /* Hook to the break signal */
  signal(SIGINT, break_handler);

  /* initialize video & audio coders */
  pthread_create(&video_thread, NULL, video_callback, NULL);
  if(audio_fd)
    pthread_create(&audio_thread, NULL, audio_callback, NULL);

  video_frame = 0;
  audio_frame = 0;
  ea_frame = 0;
  ev_frame = 0;
  start_frame = 0;
  return(true);
}

fbool_t cleanup()
{
  /* Ignore further break signals */
  signal(SIGINT, SIG_IGN);

  /* cleanup video & audio coders */
  pthread_cond_signal(&buffer_cond);
  pthread_join(video_thread, NULL);
  pthread_cond_signal(&buffer_cond);
  if(audio_fd)
    pthread_join(audio_thread, NULL);

  /* close files */
  if(audio_fd)
    close(audio_fd);
  close(video_fd);

  return(true);
}

void break_handler(int signal)
{
  /* Handle the break (Ctrl-C) signal */
  if(signal != SIGINT) return;

  quit = true;
}


int main(int argc, char * argv[], char *env[])
{
  if(parse(argc, argv))
  {
    usage(argv[0]);
    exit(1);
  }

  if(!init())
  {
    usage(argv[0]);
    cleanup();
    exit(1);
  }

  ioctl (0, TCGETA, &ostate);/* save old state */
  ioctl (0, TCGETA, &nstate);/* get base of new state */
  nstate.c_lflag &= ~(ISIG | ICANON | ECHO);
  nstate.c_cc[VEOF] = 1;
  nstate.c_cc[VEOL] = 1;
  nstate.c_iflag &= ~(INLCR | ICRNL | IXON | IXOFF);
  nstate.c_oflag &= ~(OPOST);
  ioctl (0, TCSETA, &nstate);/* set mode */

  while(!quit) {
    /* wait mplex */
    pthread_mutex_lock(&buffer_mutex);
    pthread_cond_wait(&buffer_cond, &buffer_mutex);
    pthread_mutex_unlock(&buffer_mutex);
  }
  ioctl (0, TCSETA, &ostate);/* set mode */

  cleanup();
  
  return(0);
}
